/******************************************************************************
*
* project name:    TI-TiltMaze
* file name:       tiltmaze.c
* initial date:    29/11/2000
* authors:         thomas.nussbaumer@gmx.net
*                  marcos.lopez@wol.es
*
* description:     2-ball-tilt-maze game with 41 levels
*
* compile with: tigcc -O2 tiltmaze.c
*
*
* NOTE: !! No animals or trees were harmed during the creation this game !!
*
*
* $Id: tiltmaze.c,v 1.2 2002/11/18 08:59:44 nussbaumer Exp $
*
******************************************************************************/

#define SAVE_SCREEN
#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT

#define USE_TI89
#define USE_TI92PLUS
#define USE_V200

#include <tigcclib.h>


#define PROGRAM_NAME    "TI-TiltMaze"
#define PROGRAM_VERSION "v1.10"



short  offsetx = 0;    // used for TI92+ to center display
short  offsety = 0;    // used for TI92+ to center display

#include "tiles.c"   // tiles generated from picture with iStudio and TTTiler
#include "levels.c"  // level data


/*===========================================================================*/
/* get input key pressed by user                                             */
/*                                                                           */
/* replacement for ngetchx() and/or GKeyIn(), because:                       */
/*                                                                           */
/* (1) ngetchx() won't check the APD timer value                             */
/* (2) GKeyIn() destroys the display due to setting the status display       */
/*===========================================================================*/
short GetUserInput(void) {
    void *kbq = kbd_queue();
    short key;

    //-------------------------------------------
    // restart the APD timer now
    //-------------------------------------------
    OSTimerRestart(APD_TIMER);

    while (1) {
        //---------------------------------------
        // wait until we get a key
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        while (OSdequeue(&key, kbq)) {
            if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }
        }

        //---------------------------------------
        // handle calculator-off key combinations
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        if (key == 4363 || (TI89 && key == 16651) || (!TI89 && key == 8459)) {
            off();
            OSTimerRestart(APD_TIMER);
        }
        else {
            return key & 0xf7ff;  // mask out 0x800 (repeat flag)
        }
    }
}

#include "logo.c"  // as usual: the TICT logo screen

// the ball sprite
unsigned short ball_sprite[12] = {
0x0F00,0x3FC0,0x7FE0,0x7CE0,0xFF70,0xFF70,
0xFFF0,0xFFF0,0x7FE0,0x7FE0,0x3FC0,0x0F00
};

// the destination field sprite
unsigned short dest_sprite[12] = {
0x0F00,0x30C0,0x4020,0x50A0,0x8910,0x8610,
0x8610,0x8910,0x50A0,0x4020,0x30C0,0x0F00
};

// mask for ball and destination field
unsigned short mask[12] = {
0x000F,0x000F,0x000F,0x000F,0x000F,0x000F,
0x000F,0x000F,0x000F,0x000F,0x000F,0x000F
};

short act_level = 0; // the actual level
short moves     = 0; // number moves executed by user

typedef struct {   // simple structure for ball and destination coordinates
    short x;
    short y;
} position_t;

position_t balls[2] = {{0,0},{1,0}};
position_t dests[2] = {{MAZE_SIZE_X-2,MAZE_SIZE_Y-1},{MAZE_SIZE_X-1,MAZE_SIZE_Y-1}};


/*===========================================================================*/
/* draw a 12x12 pixel large sprite                                           */
/*===========================================================================*/
void DrawSprite(unsigned short* data,short x,short y) {
    Sprite16(x*16+3+offsetx,y*16+3+offsety,12,mask,LCD_MEM,SPRT_AND);
    Sprite16(x*16+3+offsetx,y*16+3+offsety,12,data,LCD_MEM,SPRT_OR);
}


/*===========================================================================*/
/* animate ball moving from (x1,y1) to (x2,y2)                               */
/*===========================================================================*/
void MoveBall(short x1,short y1,short x2,short y2) {
    short redraw1 = 0;
    short redraw2 = 0;

    if (x1 == dests[0].x && y1 == dests[0].y) redraw1 = 1;
    if (x1 == dests[1].x && y1 == dests[1].y) redraw2 = 1;

    x1 = (x1<<4)+3+offsetx;
    y1 = (y1<<4)+3+offsety;
    x2 = (x2<<4)+3+offsetx;
    y2 = (y2<<4)+3+offsety;

    // while we haven't reached the destination position ...
    while (x1 != x2 || y1 != y2) {
        Sprite16(x1,y1,12,mask,LCD_MEM,SPRT_AND);
        if (x1 < x2) x1++;
        if (x1 > x2) x1--;
        if (y1 < y2) y1++;
        if (y1 > y2) y1--;
        if (redraw1) DrawSprite(dest_sprite,dests[0].x,dests[0].y);
        if (redraw2) DrawSprite(dest_sprite,dests[1].x,dests[1].y);
        Sprite16(x1,y1,12,ball_sprite,LCD_MEM,SPRT_OR);
    }
}

// macro to access field (x,y) of level m
#define MAZE(m,x,y)  leveldata[m][(y)*MAZE_SIZE_X+(x)]


/*===========================================================================*/
/* handle a move to the left                                                 */
/*===========================================================================*/
inline short MoveLeft(void) {
    short moved;
    short nr_moves = 0;

    do {
        moved = 0;
        if (balls[0].x < balls[1].x) {
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_LEFT)) {
                MoveBall(balls[0].x,balls[0].y,balls[0].x-1,balls[0].y);
                balls[0].x--;
                moved = 1;
                nr_moves++;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_LEFT) &&
                (balls[1].x-1 != balls[0].x || balls[1].y != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x-1,balls[1].y);
                balls[1].x--;
                moved = 1;
                nr_moves++;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_LEFT)) {
                MoveBall(balls[1].x,balls[1].y,balls[1].x-1,balls[1].y);
                balls[1].x--;
                moved = 1;
                nr_moves++;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_LEFT) &&
                (balls[0].x-1 != balls[1].x || balls[0].y != balls[1].y))
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x-1,balls[0].y);
                balls[0].x--;
                nr_moves++;
                moved = 1;
            }
        }
    }
    while (moved);

    if (nr_moves) return 1;
    return 0;
}


/*===========================================================================*/
/* handle a move to the right                                                */
/*===========================================================================*/
inline short MoveRight(void) {
    short moved;
    short nr_moves = 0;

    do {
        moved = 0;
        if (balls[0].x > balls[1].x) {
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_RIGHT)) {
                MoveBall(balls[0].x,balls[0].y,balls[0].x+1,balls[0].y);
                balls[0].x++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_RIGHT) &&
                (balls[1].x+1 != balls[0].x || balls[1].y != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x+1,balls[1].y);
                balls[1].x++;
                nr_moves++;
                moved = 1;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_RIGHT)) {
                MoveBall(balls[1].x,balls[1].y,balls[1].x+1,balls[1].y);
                balls[1].x++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_RIGHT) &&
                (balls[0].x+1 != balls[1].x || balls[0].y != balls[1].y))
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x+1,balls[0].y);
                balls[0].x++;
                nr_moves++;
                moved = 1;
            }
        }
    }
    while (moved);

    if (nr_moves) return 1;
    return 0;
}


/*===========================================================================*/
/* handle an upwards move                                                    */
/*===========================================================================*/
inline short MoveUp(void) {
    short moved;
    short nr_moves = 0;

    do {
        moved = 0;
        if (balls[0].y < balls[1].y) {
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_TOP)) {
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y-1);
                balls[0].y--;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_TOP) &&
                (balls[1].x != balls[0].x || balls[1].y-1 != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y-1);
                balls[1].y--;
                nr_moves++;
                moved = 1;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_TOP)) {
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y-1);
                balls[1].y--;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_TOP) &&
                (balls[0].x != balls[1].x || balls[0].y-1 != balls[1].y))
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y-1);
                balls[0].y--;
                nr_moves++;
                moved = 1;
            }
        }
    }
    while (moved);

    if (nr_moves) return 1;
    return 0;
}


/*===========================================================================*/
/* handle a downwards move                                                   */
/*===========================================================================*/
inline short MoveDown(void) {
    short moved;
    short nr_moves = 0;

    do {
        moved = 0;
        if (balls[0].y > balls[1].y) {
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_BOTTOM)) {
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y+1);
                balls[0].y++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_BOTTOM) &&
                (balls[1].x != balls[0].x || balls[1].y+1 != balls[0].y))
            {
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y+1);
                balls[1].y++;
                nr_moves++;
                moved = 1;
            }
        }
        else {
            if (!(MAZE(act_level,balls[1].x,balls[1].y) & MASK_WALL_BOTTOM)) {
                MoveBall(balls[1].x,balls[1].y,balls[1].x,balls[1].y+1);
                balls[1].y++;
                nr_moves++;
                moved = 1;
            }
            if (!(MAZE(act_level,balls[0].x,balls[0].y) & MASK_WALL_BOTTOM) &&
                (balls[0].x != balls[1].x || balls[0].y+1 != balls[1].y))
            {
                MoveBall(balls[0].x,balls[0].y,balls[0].x,balls[0].y+1);
                balls[0].y++;
                nr_moves++;
                moved = 1;
            }
        }
    }
    while (moved);

    if (nr_moves) return 1;
    return 0;
}


/*===========================================================================*/
/* initialize level                                                          */
/*===========================================================================*/
void InitLevel(void) {
    short x,y;
    WIN_RECT rect;

    balls[0].x   = 0;
    balls[0].y   = 0;
    balls[1].x   = 1;
    balls[1].y   = 0;

    moves = 0;

    memset(LCD_MEM,0,LCD_SIZE);

    rect.x0=offsetx;
    rect.y0=offsety;
    rect.x1=MAZE_SIZE_X*16+1+offsetx;
    rect.y1=MAZE_SIZE_Y*16+1+offsety;
    DrawClipRect(&rect, &(SCR_RECT){{0, 0, 239, 127}}, A_NORMAL);


    for (y=0;y<MAZE_SIZE_Y;y++) {
        for (x=0;x<MAZE_SIZE_X;x++) {
            short offset = leveldata[act_level][y*MAZE_SIZE_X+x]*16;
            Sprite16(x*16+1+offsetx,y*16+1+offsety,16,&tiles[offset],LCD_MEM,SPRT_OR);
        }
    }

    for (x=0;x<2;x++) {
        DrawSprite(dest_sprite,dests[x].x,dests[x].y);
        DrawSprite(ball_sprite,balls[x].x,balls[x].y);
    }

    FontSetSys(F_4x6);
    DrawStr(96+offsetx,10+offsety,"(c) TICT 2002",A_NORMAL);
    DrawStr(85+offsetx,60+offsety,"[+][-] Change Level",  A_NORMAL);
    DrawStr(85+offsetx,68+offsety,"[CLR]    Restart Level", A_NORMAL);
    DrawStr(85+offsetx,76+offsety,"[ESC]    Exit Game",     A_NORMAL);
    FontSetSys(F_6x8);
    DrawStr(86+offsetx,offsety,PROGRAM_NAME,A_REPLACE);
    DrawStr(87+offsetx,offsety,PROGRAM_NAME,A_NORMAL);
    printf_xy(94+offsetx,25+offsety,"Level: %d",act_level+1);

}


/*===========================================================================*/
/* checks if both balls are on the destination fields                        */
/*===========================================================================*/
short CheckWin(void) {
    if ((balls[0].x == dests[0].x && balls[0].y == dests[0].y) ||
        (balls[0].x == dests[1].x && balls[0].y == dests[1].y))
    {
        if ((balls[1].x == dests[0].x && balls[1].y == dests[0].y) ||
            (balls[1].x == dests[1].x && balls[1].y == dests[1].y))
        {
            return 1;
        }
    }

    return 0;
}


/*===========================================================================*/
/* ... as usual: our main routine                                            */
/*===========================================================================*/
void _main(void) {
    short input;

    //---------------------------------------------
    // setup offsetx/offsety depending on calc type
    //---------------------------------------------
    if (!TI89) offsetx = 40,offsety = 14;
    else       offsetx = offsety = 0;

    //--------------------------------------------------------------------
    // turn grayscale on, display logo screen and turn grayscale off again
    //--------------------------------------------------------------------
    if (!GrayOn()) return;
    Logo();
    GrayOff();

    //-------------------------------------
    // initialize level and enter main loop
    //-------------------------------------
    InitLevel();

    do {
        //-------------------------------------------------------
        // update number of moves display and wait for user input
        //-------------------------------------------------------
        printf_xy(94+offsetx,35+offsety,"Moves: %d",moves);
        input = GetUserInput();

        //-----------------------------------------
        // handle cursor keys and restart level key
        //-----------------------------------------
        if (input == KEY_LEFT)       moves += MoveLeft();
        else if (input == KEY_RIGHT) moves += MoveRight();
        else if (input == KEY_UP)    moves += MoveUp();
        else if (input == KEY_DOWN)  moves += MoveDown();
        else if (input == KEY_CLEAR) InitLevel();

        //--------------------------------------------------
        // if user has solved the level, display message and
        // proceed to next level
        //--------------------------------------------------
        if (CheckWin()) {
            DrawStr(7+offsetx,87+offsety,"Excellent! Level solved!",A_REPLACE);
            GetUserInput();
            input = '+';  // schedule pseudo-input of '+' to increment level
        }

        //------------------------------------
        // handle level-up and level-down keys
        //------------------------------------
        if (input == '+') {
            act_level++;
            if (act_level >= MAZE_LEVELS) act_level = 0;
            InitLevel();
        }
        else if (input == '-') {
            act_level--;
            if (act_level < 0) act_level = MAZE_LEVELS-1;
            InitLevel();
        }

    }
    while (input != KEY_ESC);
}


//=============================================================================
// Revision History
//=============================================================================
//
// $Log: tiltmaze.c,v $
// Revision 1.2  2002/11/18 08:59:44  nussbaumer
// changes for V200/AMS2.08/TIGCC 0.94
//
// Revision 1.1  2000/12/09 16:04:21  Thomas Nussbaumer
// initial version
//
//
